################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.units.length import LengthUnit
from qudt.quantity import Quantity

import unittest


class LengthUnitTest(unittest.TestCase):
    def test_electron_volt(self) -> None:
        temp = Quantity(23.5, LengthUnit.NM)

        assert temp.unit is not None

        self.assertEqual("nm", temp.unit.abbreviation)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.ontology.ontology_utils import OntologyUtils

import unittest


OntologyUtils.register_namespace('qudt', 'http://qudt.org/schema/qudt#')


class OntologyUtilsTest(unittest.TestCase):
    def test_get_namespace(self) -> None:
        namespace = OntologyUtils.get_namespace('qudt')

        self.assertEqual('http://qudt.org/schema/qudt#', namespace)

    def test_get_iri(self) -> None:
        resource_iri = OntologyUtils.get_iri('qudt', 'symbol')

        self.assertEqual('http://qudt.org/schema/qudt#symbol', resource_iri)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.multiplier import Multiplier

import dataclasses


@dataclasses.dataclass
class Unit(object):
    """
    A unit of measurement.
    """
    resource_iri: str
    label: str = dataclasses.field(default_factory=str)
    abbreviation: str = dataclasses.field(default_factory=str)
    symbol: str = dataclasses.field(default_factory=str)
    type_iri: str = dataclasses.field(default_factory=str)
    multiplier: Multiplier = dataclasses.field(default_factory=Multiplier)

    def __repr__(self) -> str:
        return str(self.abbreviation)



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.unit import Unit
from qudt.ontology.unit_factory import UnitFactory

from typing import Dict
from typing import List
from typing import Optional


def long_iri(shortened_iri: str) -> str:
    if shortened_iri.startswith('uo:'):
        return 'http://purl.obolibrary.org/obo/' + shortened_iri[3:]
    elif shortened_iri.startswith('ops:'):
        return 'http://www.openphacts.org/units/' + shortened_iri[4:]
    elif shortened_iri.startswith('qudt:'):
        return 'http://qudt.org/schema/qudt#' + shortened_iri[5:]

    raise ValueError(f'Invalid shortened IRI: {shortened_iri}')


class UnitOntologyFactory(object):
    """
    A class for creating units from the Unit Ontology.
    """
    uo_to_qudt: Dict[str, str] = {
        long_iri('uo:EFO_0004374'): long_iri('ops:MilligramPerDeciliter'),
        long_iri('uo:EFO_0004385'): long_iri('ops:PicogramPerMilliliter'),
        long_iri('uo:UO_0000009'): long_iri('qudt:Kilogram'),
        long_iri('uo:UO_0000010'): long_iri('qudt:SecondTime'),
        long_iri('uo:UO_0000015'): long_iri('qudt:Centimeter'),
        long_iri('uo:UO_0000016'): long_iri('qudt:Millimeter'),
        long_iri('uo:UO_0000017'): long_iri('qudt:Micrometer'),
        long_iri('uo:UO_0000018'): long_iri('ops:Nanometer'),
        long_iri('uo:UO_0000021'): long_iri('qudt:Gram'),
        long_iri('uo:UO_0000022'): long_iri('ops:Milligram'),
        long_iri('uo:UO_0000023'): long_iri('ops:Microgram'),
        long_iri('uo:UO_0000024'): long_iri('ops:Nanogram'),
        long_iri('uo:UO_0000025'): long_iri('ops:Picogram'),
        long_iri('uo:UO_0000026'): long_iri('ops:Femtogram'),
        long_iri('uo:UO_0000027'): long_iri('qudt:DegreeCelsius'),
        long_iri('uo:UO_0000028'): long_iri('qudt:Millisecond'),
        long_iri('uo:UO_0000031'): long_iri('qudt:MinuteTime'),
        long_iri('uo:UO_0000032'): long_iri('qudt:Hour'),
        long_iri('uo:UO_0000033'): long_iri('qudt:Day'),
        long_iri('uo:UO_0000039'): long_iri('qudt:Micromole'),
        long_iri('uo:UO_0000040'): long_iri('qudt:Millimole'),
        long_iri('uo:UO_0000041'): long_iri('qudt:Nanomole'),
        long_iri('uo:UO_0000042'): long_iri('qudt:Picomole'),
        long_iri('uo:UO_0000043'): long_iri('qudt:Femtomole'),
        long_iri('uo:UO_0000062'): long_iri('ops:Molar'),
        long_iri('uo:UO_0000063'): long_iri('ops:Millimolar'),
        long_iri('uo:UO_0000064'): long_iri('ops:Micromolar'),
        long_iri('uo:UO_0000065'): long_iri('ops:Nanomolar'),
        long_iri('uo:UO_0000066'): long_iri('ops:Picomolar'),
        long_iri('uo:UO_0000073'): long_iri('ops:Femtomolar'),
        long_iri('uo:UO_0000098'): long_iri('ops:Milliliter'),
        long_iri('uo:UO_0000099'): long_iri('qudt:Liter'),
        long_iri('uo:UO_0000101'): long_iri('ops:Microliter'),
        long_iri('uo:UO_0000169'): long_iri('ops:PartsPerMillion'),
        long_iri('uo:UO_0000173'): long_iri('ops:GramPerMilliliter'),
        long_iri('uo:UO_0000175'): long_iri('ops:GramPerLiter'),
        long_iri('uo:UO_0000176'): long_iri('ops:MilligramPerMilliliter'),
        long_iri('uo:UO_0000187'): long_iri('qudt:Percent'),
        long_iri('uo:UO_0000197'): long_iri('ops:LiterPerKilogram'),
        long_iri('uo:UO_0000198'): long_iri('ops:MilliliterPerKilogram'),
        long_iri('uo:UO_0000271'): long_iri('ops:MicroliterPerMinute'),
        long_iri('uo:UO_0000272'): long_iri('qudt:MillimeterOfMercury'),
        long_iri('uo:UO_0000274'): long_iri('ops:MicrogramPerMilliliter'),
        long_iri('uo:UO_0000275'): long_iri('ops:NanogramPerMilliliter'),
        long_iri('uo:UO_0000308'): long_iri('ops:MilligramPerKilogram'),
        #long_iri('uo:UO_0000311'), longIRI(''),
    }

    # Reverse the lookup table
    qudt_to_uo: Dict[str, str] = {
        v: k for k, v in uo_to_qudt.items()
    }

    @classmethod
    def get_unit(cls, resource_iri: str) -> Optional[Unit]:
        """
        Get a unit from a Unit Ontology resource IRI.

        :param resource_iri: The IRI of a resource in the Unit Ontology
        :return: The resolved unit, or None on error
        """
        mapped_iri: Optional[str] = cls.uo_to_qudt.get(resource_iri)

        if mapped_iri:
            return UnitFactory.get_unit(mapped_iri)

        return None

    @classmethod
    def get_iris(cls, type_iri: str) -> List[str]:
        """
        Return a list of unit IRIs with the given unit type.

        :param type_iri: The IRI of the unit type, e.g. 'http://qudt.org/schema/qudt#MolarConcentrationUnit'
        :return: The list of IRIs, or empty if no units match the specified type
        """
        iris: List[str] = list()

        qudt_iris: List[str] = UnitFactory.get_iris(type_iri)

        for qudt_iri in qudt_iris:
            uo_iri = cls.qudt_to_uo.get(qudt_iri)
            if uo_iri:
                iris.append(uo_iri)

        return iris



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

import json
import os
import pyld.jsonld
import rdflib
from typing import Optional


class OntologyReader(object):
    """
    Class to read an RDF triplet ontology repository.
    """

    @classmethod
    def read(cls, repo_path: str) -> rdflib.Graph:
        """
        Read an RDF triplet repository.

        :param repo_path: The path to the RDF repository
        :return: The RDF graph, or None on error
        """
        # Get the repo format based on the path
        repo_format = cls._get_repo_format(repo_path)

        # Use a conjunctive graph to fix loading the QUDT JSON-LD repository
        # See https://github.com/RDFLib/rdflib-jsonld/issues/53
        # TODO: Fix URL
        g = rdflib.ConjunctiveGraph()

        if repo_format == 'json-ld':
            # JSON-LD support in rdflib is limited. Particularly, while it can
            # handle namespaces, it cannot handle fully compacted JSON-LD.
            #
            # To allow for compacted JSON-LD, we use pyld to expand the JSON-LD
            # for rdflib.
            with open(repo_path, 'r') as file:
                compacted = json.loads(file.read())
            expanded = pyld.jsonld.expand(compacted)
            g.parse(data=json.dumps(expanded), format=repo_format)
        else:
            g.parse(repo_path, format=repo_format)

        return g

    @staticmethod
    def _get_repo_format(repo_path: str) -> Optional[str]:
        """
        Get the RDF format used by RDFLib based on the file's extension.

        :param repo_path: The path to the RDF repository
        :return: The repo format, or None if unknown (assumed XML by RDFLib)
        """
        _, repo_ext = os.path.splitext(repo_path)

        formats = {
            '.jsonld': 'json-ld',
            '.ttl': 'turtle'
        }

        return formats.get(repo_ext)



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.unit import Unit

import dataclasses
from typing import Optional


@dataclasses.dataclass
class Quantity(object):
    """
    A quantity with a value and a unit.
    """
    value: float
    unit: Optional[Unit]

    def convert_to(self, unit: Unit) -> 'Quantity':
        """
        Converts the quantity's value to the specified unit of measurement.

        :param unit: The target unit
        :return: The converted quantity
        """
        if not unit:
            raise ValueError('Target unit cannot be null')

        if not self.unit:
            raise ValueError('This measurement does not have units defined')

        if self.unit == unit:
            # Nothing to be done
            return self

        if self.unit.type_iri != unit.type_iri:
            raise ValueError(
                f'The new unit does not have the same parent type '
                f'(source: {self.unit.type_iri}; target: {unit.type_iri})'
            )

        # Convert to the base unit
        base_unit_value = self.value * self.unit.multiplier.multiplier + self.unit.multiplier.offset

        # Convert the base unit to the new unit
        new_value = (base_unit_value - unit.multiplier.offset) / unit.multiplier.multiplier

        new_measurement = Quantity(
            unit=unit,
            value=new_value,
        )

        return new_measurement

    def __repr__(self) -> str:
        """
        Return a string representation of the quantity.
        """
        return f'{self.value} {self.unit}'



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

import dataclasses


@dataclasses.dataclass
class Multiplier(object):
    """
    A multiplier with an optional offset.
    """
    offset: float = dataclasses.field(default=0.0)
    multiplier: float = dataclasses.field(default=1.0)



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.units.temperature import TemperatureUnit
from qudt.quantity import Quantity

import unittest


class CelsiusTest(unittest.TestCase):
    def test_absolute_zero(self) -> None:
        temp = Quantity(-273.15, TemperatureUnit.CELSIUS)
        temp2 = temp.convert_to(TemperatureUnit.KELVIN)

        self.assertEqual(TemperatureUnit.KELVIN, temp2.unit)
        self.assertAlmostEqual(0.0, temp2.value)

    def test_room_temperature(self) -> None:
        temp = Quantity(20, TemperatureUnit.CELSIUS)
        temp2 = temp.convert_to(TemperatureUnit.KELVIN)

        self.assertEqual(TemperatureUnit.KELVIN, temp2.unit)
        self.assertAlmostEqual(293.15, temp2.value)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from .mass_per_volume_unit_test import MassPerVolumeUnitTest
from .molar_test import MolarTest
from .nanomolar_test import NanomolarTest



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.unit import Unit
from qudt.ontology.unit_factory import UnitFactory


class ConcentrationUnit(object):
    """
    """
    MOLE_PER_CUBIC_METER: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#MolePerCubicMeter')

    MOLAR: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Molar')
    MILLIMOLAR: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Millimolar')
    NANOMOLAR: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Nanomolar')
    MICROMOLAR: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Micromolar')

    GRAM_PER_LITER: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/GramPerLiter')
    MICROGRAM_PER_MILLILITER: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/MicrogramPerMilliliter')
    PICOGRAM_PER_MILLILITER: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/PicogramPerMilliliter')



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.ontology.unit_factory import UnitFactory
from qudt.unit import Unit

import unittest


class UnitFactoryTest(unittest.TestCase):
    def test_get_instance(self) -> None:
        factory = UnitFactory._get_instance()

        self.assertTrue(factory)

    def test_get_unit(self) -> None:
        unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#Kelvin')

        self.assertTrue(isinstance(unit, Unit))
        self.assertEqual('Kelvin', unit.label)
        self.assertEqual('K', unit.symbol)
        self.assertEqual('K', unit.abbreviation)
        self.assertEqual(1, unit.multiplier.multiplier)
        self.assertEqual(0, unit.multiplier.offset)
        self.assertEqual('http://qudt.org/schema/qudt#TemperatureUnit', unit.type_iri)

    def test_get_iris(self) -> None:
        units = UnitFactory.get_iris('http://qudt.org/schema/qudt#TemperatureUnit')

        self.assertTrue(units)
        self.assertGreaterEqual(len(units), 1)

    def test_get_open_phacts_unit(self) -> None:
        unit = UnitFactory.get_unit('http://www.openphacts.org/units/Nanomolar')

        self.assertTrue(unit)
        self.assertEqual('Nanomolar', unit.label)
        self.assertEqual('nmol/dm^3', unit.symbol)
        self.assertEqual('nM', unit.abbreviation)
        self.assertAlmostEqual(0.000001, unit.multiplier.multiplier)
        self.assertAlmostEqual(0, unit.multiplier.offset)
        self.assertEqual('http://qudt.org/schema/qudt#MolarConcentrationUnit', unit.type_iri)

    def test_get_open_phacts_unit_newer(self) -> None:
        unit = UnitFactory.get_unit('http://www.openphacts.org/units/NanogramPerMilliliter')

        self.assertTrue(unit)
        self.assertEqual('http://qudt.org/schema/qudt#MassPerVolumeUnit', unit.type_iri)

    def test_find_units(self) -> None:
        units = UnitFactory.find_units('nM')

        self.assertTrue(units)
        self.assertGreaterEqual(len(units), 1)
        self.assertEqual('http://www.openphacts.org/units/Nanomolar', units[0].resource_iri)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.ontology.ontology_reader import OntologyReader
from qudt.ontology.unit_factory import UnitFactory

import os
import unittest

ONTOLOGY_FILE = 'ops.ttl'


class OntologyReaderTest(unittest.TestCase):
    def test_unit_ontology(self) -> None:
        schema_path = UnitFactory.get_repo_dir()

        repo_path = os.path.join(schema_path, ONTOLOGY_FILE)

        repos = OntologyReader.read(repo_path)

        self.assertGreaterEqual(len(repos), 1)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.uo.unit_ontology_factory import UnitOntologyFactory
from qudt.unit import Unit

import unittest


class UnitOntologyFactoryTest(unittest.TestCase):
    def test_get_unit(self) -> None:
        unit = UnitOntologyFactory.get_unit('http://purl.obolibrary.org/obo/UO_0000065')

        assert unit is not None

        self.assertTrue(isinstance(unit, Unit))
        self.assertEqual('Nanomolar', unit.label)
        self.assertEqual('nmol/dm^3', unit.symbol)
        self.assertEqual('nM', unit.abbreviation)
        self.assertEqual(0.000001, unit.multiplier.multiplier)
        self.assertEqual(0, unit.multiplier.offset)
        self.assertEqual('http://qudt.org/schema/qudt#MolarConcentrationUnit', unit.type_iri)

    def test_get_units_by_qudt_type(self) -> None:
        units = UnitOntologyFactory.get_iris('http://qudt.org/schema/qudt#MolarConcentrationUnit')

        self.assertTrue(units)
        self.assertGreaterEqual(len(units), 1)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.units.concentration import ConcentrationUnit
from qudt.quantity import Quantity

import unittest


class NanomolarTest(unittest.TestCase):
    def test_molar_conversion(self) -> None:
        obs = Quantity(0.1, ConcentrationUnit.MICROMOLAR)
        obs2 = obs.convert_to(ConcentrationUnit.NANOMOLAR)

        self.assertEqual(ConcentrationUnit.NANOMOLAR, obs2.unit)
        self.assertAlmostEqual(100, obs2.value)

    def test_compare_to_mole_per_cubic_meter(self) -> None:
        obs = Quantity(1.0, ConcentrationUnit.NANOMOLAR)
        obs2 = obs.convert_to(ConcentrationUnit.MOLE_PER_CUBIC_METER)

        self.assertEqual(ConcentrationUnit.MOLE_PER_CUBIC_METER, obs2.unit)
        self.assertAlmostEqual(1.0e-6, obs2.value)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.ontology.ontology_reader import OntologyReader
from qudt.ontology.qudt import QUDT
from qudt.ontology.rdf import RDF
from qudt.ontology.rdfs import RDFS
from qudt.unit import Unit

import os
import rdflib
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple


# Type definitions
Statement = Tuple[str, str, str]
Predicate = Callable[[str, str, str], bool]


# The package containing the RDF triplet repositories
REPO_PACKAGE_NAME = 'resources'

# The RDF triplet repositories to load
REPO_FILES = [
    'openphacts.jsonld',
    'unit.jsonld',
]


class UnitFactory(object):
    """
    A factory for creating units of measurement.
    """

    _instance: Optional['UnitFactory'] = None

    def __init__(self):
        """
        Create an instance of the unit factory and load the RDF triplet
        repositories.
        """
        # Get the path to this package
        package_path = os.path.dirname(os.path.realpath(__file__))

        # Get the path to the repository files
        self._repo_path: str = os.path.join(package_path, REPO_PACKAGE_NAME)

        # Load the repositories
        self._repos: List[rdflib.Graph] = [
            self._read_repo(repo_file) for repo_file in REPO_FILES
        ]

    @classmethod
    def _get_instance(cls) -> 'UnitFactory':
        """
        Get the singleton used to store repository contents.

        :return: The singleton instance of type UnitFactory
        """
        if not cls._instance:
            cls._instance = UnitFactory()

        return cls._instance

    @classmethod
    def get_repo_dir(cls) -> str:
        """
        Get the directory used for loading repostories.

        :return: The path to the repository directory
        """
        return cls._get_instance()._repo_path

    @classmethod
    def get_unit(cls, resource_iri: str) -> Unit:
        """
        Get a unit by its resource IRI.

        :param resource_iri: The unit's resource IRI
        :return: The unit, or None on error
        """
        return cls._get_instance()._get_unit(resource_iri)

    def _get_unit(self, resource_iri: str) -> Unit:
        """
        Internal implementation of get_unit().
        """
        unit = Unit(
            resource_iri=resource_iri,
        )

        statements: List[Statement] = self._get_statements(
            self._repos,
            lambda subj, pred, obj: str(subj) == resource_iri,
        )

        for (subject, predicate, obj) in statements:
            if predicate == QUDT.SYMBOL:
                unit.symbol = str(obj)
            elif predicate == QUDT.ABBREVIATION:
                unit.abbreviation = str(obj)
            elif predicate == QUDT.CONVERSION_OFFSET:
                unit.multiplier.offset = float(obj)
            elif predicate == QUDT.CONVERSION_MULTIPLIER:
                unit.multiplier.multiplier = float(obj)
            elif predicate == RDFS.LABEL:
                unit.label = str(obj)
            elif predicate == RDF.TYPE:
                type_iri = str(obj)
                if not self._should_be_ignored(type_iri):
                    unit.type_iri = type_iri

        return unit

    @classmethod
    def find_units(cls, abbreviation: str) -> List[Unit]:
        """
        Get units by their abbreviation.

        :param abbreviation: The unit abbreviation, e.g. 'nM'
        :return: The list of units, or empty if no units matched the symbol
        """
        return cls._get_instance()._find_units(abbreviation)

    def _find_units(self, abbreviation: str) -> List[Unit]:
        """
        Internal implementation of find_units()
        """
        found_units: List[Unit] = list()

        statements: List[Statement] = self._get_statements(
            self._repos,
            lambda subj, pred, o: str(pred) == QUDT.ABBREVIATION and str(o) == abbreviation,
        )

        for (subject, predicate, obj) in statements:
            type_iri = subject
            found_units.append(self._get_unit(type_iri))

        return found_units

    @classmethod
    def get_iris(cls, type_iri: str) -> List[str]:
        """
        Return a list of unit IRIs with the given unit type.

        :param type_iri: The IRI of the unit type, e.g. 'http://qudt.org/schema/qudt#TemperatureUnit'
        :return: The list of units, or empty if none match the specified type
        """
        return cls._get_instance()._get_iris(type_iri)

    def _get_iris(self, type_iri: str) -> List[str]:
        """
        Internal implementation of get_iris()
        """
        statements: List[Statement] = self._get_statements(
            self._repos,
            lambda subj, pred, o: str(o) == type_iri,
        )

        return [subj for (subj, pred, o) in statements]

    def _read_repo(self, file_name: str) -> rdflib.Graph:
        """
        Helper function to load the RDF triplet repository.

        :param file_name: The path to the repo
        :return: The loaded graph object
        """
        repo_path = os.path.join(self._repo_path, file_name)

        return OntologyReader.read(repo_path)

    @staticmethod
    def _get_statements(
            repos: List[rdflib.Graph],
            triplet_test: Predicate
    ) -> List[Statement]:
        """
        Get the statements of the given repos that satisfy the provided lambda.

        :param repos: The ontology repositories
        :param triplet_test: The lambda to invoke per statement
        :return: The matching statements
        """
        statements: List[Statement] = list()

        for repo in repos:
            for (subject, predicate, obj) in repo:
                if triplet_test(subject, predicate, obj):
                    statements.append((str(subject), str(predicate), obj))

        return statements

    @staticmethod
    def _should_be_ignored(type_iri: str) -> bool:
        """
        Check if a statement should be ignored when constructing units.

        :param type_iri: The predicate type
        :return: True if the statement should be ignored, False otherwise
        """
        # Accept anything outside the QUDT namespace
        if not type_iri.startswith(QUDT.namespace):
            return False

        if type_iri in [
            QUDT.SI_DERIVED_UNIT,
            QUDT.SI_BASE_UNIT,
            QUDT.SI_UNIT,
            QUDT.DERIVED_UNIT,
            QUDT.NOT_USED_WITH_SI_UNIT,
            QUDT.USED_WITH_SI_UNIT,
        ]:
            return True

        # Everything else is fine too
        return False



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#################################################################################

from qudt.ontology.ontology_utils import OntologyUtils


OntologyUtils.register_namespace('qudt', 'http://qudt.org/schema/qudt#')


class QUDT(object):
    namespace = OntologyUtils.get_namespace('qudt')

    SYMBOL = OntologyUtils.get_iri('qudt', 'symbol')
    ABBREVIATION = OntologyUtils.get_iri('qudt', 'abbreviation')
    CONVERSION_OFFSET = OntologyUtils.get_iri('qudt', 'conversionOffset')
    CONVERSION_MULTIPLIER = OntologyUtils.get_iri('qudt', 'conversionMultiplier')

    SI_UNIT = OntologyUtils.get_iri('qudt', 'SIUnit')
    SI_BASE_UNIT = OntologyUtils.get_iri('qudt', 'SIBaseUnit')
    SI_DERIVED_UNIT = OntologyUtils.get_iri('qudt', 'SIDerivedUnit')
    DERIVED_UNIT = OntologyUtils.get_iri('qudt', 'DerivedUnit')
    NOT_USED_WITH_SI_UNIT = OntologyUtils.get_iri('qudt', 'NotUsedWithSIUnit')
    USED_WITH_SI_UNIT = OntologyUtils.get_iri('qudt', 'UsedWithSIUnit')



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.units.concentration import ConcentrationUnit
from qudt.quantity import Quantity

import unittest


class MassPerVolumeUnitTest(unittest.TestCase):
    def test_electron_volt(self) -> None:
        obs = Quantity(0.1, ConcentrationUnit.MICROGRAM_PER_MILLILITER)
        obs2 = obs.convert_to(ConcentrationUnit.PICOGRAM_PER_MILLILITER)

        self.assertEqual(ConcentrationUnit.PICOGRAM_PER_MILLILITER, obs2.unit)
        self.assertAlmostEqual(100000, obs2.value)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

import setuptools


with open('README.md') as file:
    long_description = file.read()


setuptools.setup(
    name='pyqudt',
    version='1.1.0',
    author='Garrett Brown',
    author_email='themagnificentmrb@gmail.com',
    description='Python library for working with the QUDT (Quantity, Unit, Dimension and Type) ontology.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/eigendude/pyqudt',
    license='BSD-3-Clause',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    packages=setuptools.find_packages(exclude=['test']),
    package_data={
        'qudt.ontology.resources': ['*'],
    },
    install_requires=[
        'PyLD',
        'rdflib',
        'rdflib-jsonld',
    ],
)



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.ontology.qudt import QUDT

import unittest


class QUDTTest(unittest.TestCase):
    def test_unit_ontology(self) -> None:
        symbol_iri = QUDT.SYMBOL

        self.assertTrue(symbol_iri.startswith(QUDT.namespace))


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.unit import Unit
from qudt.ontology.unit_factory import UnitFactory


class TemperatureUnit(object):
    """
    """
    KELVIN: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#Kelvin')
    CELSIUS: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#DegreeCelsius')
    FAHRENHEIT: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#DegreeFahrenheit')



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from typing import Dict


class OntologyUtils(object):
    """
    Utility functions for working with ontologies.
    """

    # Registered namespaces
    _namespaces: Dict[str, str] = dict()

    @classmethod
    def register_namespace(cls, shorthand: str, namespace: str) -> None:
        """
        Register shorthand for a namespace.

        :param shorthand: The shorthand, e.g. 'xsd'
        :param namespace The namespace, e.g. 'http://www.w3.org/2001/XMLSchema#'
        """
        cls._namespaces[shorthand] = namespace

    @classmethod
    def get_namespace(cls, shorthand: str) -> str:
        """
        Get the namespace for the given shorthand.

        :param shorthand: The shorthand, e.g. 'xsd'
        :return: The namespace, e.g. 'http://www.w3.org/2001/XMLSchema#'
        """
        return cls._namespaces[shorthand]

    @classmethod
    def get_iri(cls, shorthand: str, local_part: str) -> str:
        """
        Get the full IRI given the namespace shorthand and the local part

        :param shorthand: The namespace shorthand, e.g. 'xsd'
        :param local_part: The local part of the IRI
        :return: The full IRI containing the namespace and local part
        """
        return cls.get_namespace(shorthand) + local_part



################################################################################
#
#  Copyright (C) 2019 Garrett Brown
#  This file is part of pyqudt - https://github.com/eigendude/pyqudt
#
#  pyqudt is derived from jQUDT
#  Copyright (C) 2012-2013  Egon Willighagen <egonw@users.sf.net>
#
#  SPDX-License-Identifier: BSD-3-Clause
#  See the file LICENSE for more information.
#
################################################################################

from qudt.unit import Unit
from qudt.ontology.unit_factory import UnitFactory


class MassUnit(object):
    """
    """
    KILOGRAM: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#Kilogram')
    GRAM: Unit = UnitFactory.get_unit('http://qudt.org/vocab/unit#Gram')
    MILLIGRAM: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Milligram')
    MICROGRAM: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Microgram')
    NANOGRAM: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Nanogram')
    PICOGRAM: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Picogram')
    FEMTOGRAM: Unit = UnitFactory.get_unit('http://www.openphacts.org/units/Femtogram')
